/*
 * Example Code for the Paper
 *
 * "Mesh Saliency via Local Curvature Entropy", Eurographics 2016
 *
 * by M. Limper, A. Kuijper and D. Fellner
 *
 */

#include "curvature.h"

#include <limits>
#include <cmath>
#include <float.h>
#include <map>

#include "utils.h"

//------------------------------------------------------------------------------------------------------------------

float doubleFaceArea(const OpenMesh::Vec3f & e0, const OpenMesh::Vec3f & e2)
{
     return OpenMesh::cross(e0, -e2).norm();
}

//------------------------------------------------------------------------------------------------------------------

namespace  lce
{

//------------------------------------------------------------------------------------------------------------------

bool computeCurvature(TriMesh & mesh,
                      OpenMesh::VPropHandleT<float> & gaussianCurvatureProperty,
                      OpenMesh::VPropHandleT<float> & meanCurvatureProperty     )
{
    mesh.request_face_normals();
    mesh.request_vertex_normals();
    mesh.update_face_normals();
    mesh.update_vertex_normals();


    mesh.add_property(gaussianCurvatureProperty);
    mesh.add_property(meanCurvatureProperty);


    TriMesh::VertexIter vIt, vEnd;


    //initialize properties and helpers

    std::map<TriMesh::VertexHandle, float> areaWeight;
    std::map<TriMesh::VertexHandle, OpenMesh::Vec3f> accDiffVec;

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        mesh.property(gaussianCurvatureProperty, *vIt) = M_PI * 2.0f;
        mesh.property(meanCurvatureProperty, *vIt)     = 0.0f;

        areaWeight[*vIt] = 0.0f;
        accDiffVec[*vIt] = OpenMesh::Vec3f(0.0f, 0.0f, 0.0f);
    }


    //compute vertex weights via face areas

    TriMesh::FaceIter fIt, fEnd;
    TriMesh::FaceVertexIter vfIt;

    OpenMesh::Vec3f p0, p1, p2,
                    e0, e1, e2;

    TriMesh::VertexHandle v0, v1, v2;

    float l0, l1, l2;             //squared edge lengths
    float alpha0, alpha1, alpha2; //angles
    float vA0, vA1, vA2;          //areas

    float oneByTanAlpha0, oneByTanAlpha1, oneByTanAlpha2;
    float obtA;

    float k;

    const float PiHalf   = M_PI * 0.5f;
    const float AngleEps = 1.0e-15;

    for (fIt = mesh.faces_sbegin(), fEnd = mesh.faces_end(); fIt != fEnd; ++fIt)
    {
        vfIt = mesh.fv_iter(*fIt);

        v0 = *vfIt++;
        v1 = *vfIt++;
        v2 = *vfIt;

        p0 = mesh.point(v0);
        p1 = mesh.point(v1);
        p2 = mesh.point(v2);

        e0 = p1 - p0;
        e1 = p2 - p1;
        e2 = p0 - p2;

        alpha0 = angle( e0, -e2);
        alpha1 = angle( e1, -e0);
        alpha2 = M_PI - (alpha0 + alpha1);

        if (alpha0 < AngleEps || alpha1 < AngleEps || alpha2 < AngleEps)
        {
            continue;
        }

        //accumulate vertex weights

        //obtuse cases
        if (alpha0 >= PiHalf)
        {
            obtA = doubleFaceArea(e0, e2) * 0.25f;

            areaWeight[v0] += obtA;
            areaWeight[v1] += obtA * 0.5f;
            areaWeight[v2] += obtA * 0.5f;
        }
        else if (alpha1 >= PiHalf)
        {
            obtA = doubleFaceArea(e0, e2) * 0.25f;

            areaWeight[v0] += obtA * 0.5f;
            areaWeight[v1] += obtA;
            areaWeight[v2] += obtA * 0.5f;
        }
        else if (alpha2 >= PiHalf)
        {
            obtA = doubleFaceArea(e0, e2) * 0.25f;

            areaWeight[v0] += obtA * 0.5f;
            areaWeight[v1] += obtA * 0.5f;
            areaWeight[v2] += obtA;
        }
        //non-obtuse case
        else
        {
            l0 = e0.sqrnorm();
            l1 = e1.sqrnorm();
            l2 = e2.sqrnorm();

            oneByTanAlpha0 = 1.0f / tan(alpha0);
            oneByTanAlpha1 = 1.0f / tan(alpha1);
            oneByTanAlpha2 = 1.0f / tan(alpha2);

            vA0 = (l2*oneByTanAlpha1 + l0*oneByTanAlpha2) / 8.0f;
            vA1 = (l0*oneByTanAlpha2 + l1*oneByTanAlpha0) / 8.0f;
            vA2 = (l1*oneByTanAlpha0 + l2*oneByTanAlpha1) / 8.0f;

            areaWeight[v0] += vA0;
            areaWeight[v1] += vA1;
            areaWeight[v2] += vA2;
        }


        //accumulate difference vectors

        if(alpha0 != 0.0f && alpha1 != 0.0f && alpha2 != 0.0f)
        {
            oneByTanAlpha0 = 1.0f / tan(alpha0);
            oneByTanAlpha1 = 1.0f / tan(alpha1);
            oneByTanAlpha2 = 1.0f / tan(alpha2);

            accDiffVec[v0] += (e2*oneByTanAlpha1 - e0*oneByTanAlpha2) * 0.25f;
            accDiffVec[v1] += (e0*oneByTanAlpha2 - e1*oneByTanAlpha0) * 0.25f;
            accDiffVec[v2] += (e1*oneByTanAlpha0 - e2*oneByTanAlpha1) * 0.25f;


            //accumulate values for gaussian curvature

            mesh.property(gaussianCurvatureProperty, v0) -= alpha0;
            mesh.property(gaussianCurvatureProperty, v1) -= alpha1;
            mesh.property(gaussianCurvatureProperty, v2) -= alpha2;

            //TODO: add case for adjusting gaussian curvature at borders
        }
    }


    //finally, compute final curvature values for vertices

    for (vIt = mesh.vertices_begin(), vEnd = mesh.vertices_end(); vIt != vEnd; ++vIt)
    {
        if (areaWeight[*vIt] <= std::numeric_limits<float>::epsilon())
        {
            mesh.property(gaussianCurvatureProperty, *vIt) = 0.0f;
            mesh.property(meanCurvatureProperty,     *vIt) = 0.0f;
        }
        else
        {
            mesh.property(gaussianCurvatureProperty, *vIt) /= areaWeight[*vIt];

            k = (OpenMesh::dot(accDiffVec[*vIt], mesh.normal(*vIt)) <= 0.0f ? -1.0f : 1.0f) *
                (accDiffVec[*vIt] / areaWeight[*vIt]).norm();

            mesh.property(meanCurvatureProperty,*vIt) = k;
        }
    }


    return true;
}

//------------------------------------------------------------------------------------------------------------------

} //namespace lce
